<div class="globaltoc">
{{ toctree(maxdepth=theme_navigation_depth|int,
    collapse=theme_collapse_navigation|tobool,
    includehidden=theme_includehidden|tobool,
    titles_only=theme_titles_only|tobool) }}
</div>