.dataframe {
  width: 100%;
  max-width: 100%;
  margin-bottom: 1rem;
  background-color: transparent;
  border-collapse: collapse;
  border: 0;
  display: block;
  overflow-x: scroll;
  font-size: 12px;
}

.dataframe th,
.dataframe td {
  padding: 0.75rem;
  vertical-align: top;
  border-top: 1px solid #dee2e6;
}

.dateframe thead th {
  vertical-align: bottom;
  border-bottom: 2px solid #dee2e6;
}

.dataframe tbody + tbody {
  border-top: 2px solid #dee2e6;
}

#searchbox {
  padding-top: 20px;
}

.yue img {
  max-width: 700px;
}

@media (max-width: 1200px) {
  .yue img {
    max-width: 100%;
  }
}

p.caption {
  font-weight: 700;
  font-size: 1.42em;
  margin: 1.42em 0 .4em;
}

.globaltoc>p.caption {
  display: block;
  font-size: 16px;
}

.t-sidebar>li.toctree-l1>ul {
  font-size: 80%;
}

li.current>a {
  color: var(--color-primary);
}

.versions {
  position: fixed;
  bottom: 0;
  right: 0;
  background: #f9f9f9;
  border: 1px solid #efefef;
  padding: 5px 10px;
  margin:  10px;
  font-size: 80%;
}

.current-version {
  font-weight: bold;
  cursor: pointer;
}

.current-version span {
  padding-left: 5px;
}

.versions-dropdown {
  display: none;
}

.versions dl {
  margin-block-start: 0;
  margin-block-end: 0;
  padding-top: 2px;
}

.versions dd {
  margin: 0;
}
